"The base of the extension."

from interactions.ext import Base, Version, VersionAuthor
from .persistence import Persistence

version = Version(
    version="2.2.0",
    author=VersionAuthor(
        name="Dworv",
        email="dwarvyt@gmail.com",
    ),
)

base = Base(
    name="Persistence",
    version=version,
    link="https://github.com/dworv/interactions-persistence",
    description="An extension to add simple custom_id encoding to interactions.py.",
    packages="interactions.ext.persistence",
)

def setup(bot, cipher_key=None):
    return Persistence(bot, cipher_key)



"""Here lies my attempt to not pass the bot arg in PersistentCustomID"""


# from typing import Optional, Union, List
# from interactions.client.context import CommandContext, ComponentContext, MISSING, ActionRow, Message

# from .parse import PersistentCustomID

# def hooked_str(obj):
#     class HookedStr(str):
#         def __init__(self, arg):
#             self._persistence = obj
#             super().__init__()
#     return HookedStr

# og_command_send = CommandContext.send
# og_component_send = ComponentContext.send

# def fix_component(cipher, component):
#     if isinstance(component.custom_id, PersistentCustomID):
#         component.custom_id = component.custom_id.encrypt(cipher)
#     return component

# def fix_action_row(cipher, action_row):
#     for i, component in enumerate(action_row.components):
#         if hasattr(component, "custom_id"):
#             component = fix_component(cipher, component)
#             action_row.components[i] = component
#     return action_row

# def fix_components(ctx, components):
#     cipher = ctx.member._client.token._persistence._cipher
#     if hasattr(components, "custom_id"):
#         return fix_component(cipher, components)
#     if isinstance(components, ActionRow):
#         return fix_action_row(cipher, components)
#     # if its a list of...
#     if isinstance(components, list):
#         if len(components) > 0:
#             for i, component in enumerate(components):
#                 if hasattr(components, "custom_id"):
#                     components[i] = fix_component(cipher, components)
#                 if isinstance(components, ActionRow):
#                     components[i] = fix_action_row(cipher, components)

# async def hooked_command_send(self: CommandContext, content: Optional[str] = MISSING, **kwargs) -> Message:
#     if kwargs.get("components"):
#         print("it had components")
#         kwargs["components"] = fix_components(self, kwargs["components"])

#     return await og_command_send(self, content, **kwargs)


# async def hooked_component_send(self: ComponentContext, content: Optional[str] = MISSING, **kwargs) -> Message:
#     if kwargs.get("components"):
#         kwargs["components"] = fix_components(self, kwargs["components"])
#     return await og_command_send(self, content, **kwargs)



from .base import base, version, setup
from .cipher import Cipher
from .extension import (
    PersistenceExtension,
    extension_persistent_component,
    extension_persistent_modal,
)
from .parse import PersistentCustomID, ParseError
from .persistence import Persistence
from .scripts import keygen



"""Added methods for the `interactions.Client` instance for ease of use."""

def persistent_component(self, tag: str):
    """
    The persistent component decorator for the Client object.

    Parameters:
        tag (str): The tag to identify your component.
    """

    def inner(coro):
        self.persistence.component(tag)(coro)

    return inner


def persistent_modal(self, tag: str, use_kwargs: bool = False):
    """
    The persistent modal decorator for the Client object.

    Parameters:
        tag (str): The tag to identify your modal.
        use_kwargs (bool): Whether to return key word arguments mapped with the custom_ids of the individual text inputs. Not recommended.
                (defaults to False)
    """

    def inner(coro):
        self.persistence.modal(tag, use_kwargs)(coro)

    return inner



"""The main file containing the persistence extension."""

from json import JSONDecodeError
import logging
from types import MethodType
from typing import Callable, Dict, Tuple, Union

from interactions import Extension, CommandContext, ComponentContext, extension_listener

from .cipher import Cipher
from .client import persistent_component, persistent_modal
from .parse import PersistentCustomID


class Persistence(Extension):
    """
    The Persistence extension.

    The Persistence extension handles the persistent custom_ids. It is a simple extension that makes adding information to custom_ids simple and easy.
    """

    def __init__(self, bot, cipher_key=None):
        """
        The constructor for the Persistence extension.

        Parameters:
            bot (interactions.Client): The client instance.
            cipher_key (str): The cipher key to use. When not provided, a random key will be generated and components will never persist restarts.
                (default is None)
        """
        self._cipher = Cipher(cipher_key)

        #TODO: Decide whether the following should stay typehinted or not.
        self._component_callbacks: Dict[str, Callable[[ComponentContext, Union[str, int, float, list, dict]], None]] = {}
        self._modal_callbacks: Dict[str, Tuple[Callable[[ComponentContext, Union[str, int, float, list, dict]], None], bool]] = {}

        # set as bot.persistence for convenience
        bot.persistence = self

        bot.persistent_component = MethodType(persistent_component, bot)
        bot.persistent_modal = MethodType(persistent_modal, bot)

    def component(self, tag: str):
        """
        The persistent component decorator.

        Parameters:
            tag (str): The tag to identify your component.
        """

        def inner(coro):
            self._component_callbacks[tag] = coro
            logging.debug("Registered persistent component:", tag)

        return inner

    def modal(self, tag: str, use_kwargs: bool = False):
        """
        The persistent modal decorator.

        Parameters:
            tag (str): The tag to identify your modal.
            use_kwargs (bool): Whether to return key word arguments mapped with the custom_ids of the individual text inputs. Not recommended.
                (defaults to False)
        """

        def inner(coro):
            self._modal_callbacks[tag] = (coro, use_kwargs)
            logging.debug("Registered persistent modal:", tag)

        return inner

    @extension_listener
    async def on_component(self, ctx: ComponentContext):
        """The on_component listener. This is called when a component is used."""
        if not any((
            ctx.custom_id.startswith("p~"),
            ctx.custom_id[0] == "p" and ctx.custom_id[2] == "~"
        )):
            return

        try:
            pid = PersistentCustomID.from_discord(self._cipher, ctx.custom_id)
        except JSONDecodeError:
            logging.info("Interaction made with invalid persistent custom_id. Skipping.")
        
        if pid.tag in self._component_callbacks:
            await self._component_callbacks[pid.tag](ctx, pid.package)

    @extension_listener
    async def on_modal(self, ctx: CommandContext):
        """The on_modal listener. This is called when a modal is submitted."""
        if not any((
            ctx.custom_id.startswith("p~"),
            ctx.custom_id[0] == "p" and ctx.custom_id[2] == "~"
        )):
            return

        try:
            pid = PersistentCustomID.from_discord(self._cipher, ctx.data.custom_id)
        except JSONDecodeError:
            logging.info("Interaction made with invalid persistent custom_id. Skipping.")
        
        if callback := self._modal_callbacks.get(pid.tag):
            if callback[1] == 0:
                args = [item["components"][0]["value"] for item in ctx.data.components]
                await self._modal_callbacks[pid.tag][0](ctx, pid.package, *args)
            else:
                kwargs = {item["components"][0]["custom_id"]: item["components"][0]["value"] for item in ctx.data.components}
                await self._modal_callbacks[pid.tag][0](ctx, pid.package, **kwargs)



"""The file containing the PersistentCustomID class and all of its functionality."""

from json import loads, dumps
from typing import Union, TYPE_CHECKING
from interactions import Client

from .cipher import Cipher

if TYPE_CHECKING:
    from .persistence import Persistence


def pack(obj):
    """
    Packs a compatible object into a string.

    Parameters:
        obj (object): The object to pack.
    """
    if isinstance(obj, (dict, list)):
        return dumps(obj)
    elif isinstance(obj, (str, int, float)):
        # make a list to encode it and then remove brackets
        return dumps([obj])[1:][:-1]


def unpack(encoded):
    """
    Unpacks a string into a compatible object.

    Parameters:
        encoded (str): The string to unpack.
    """
    if encoded.startswith(("{", "[")):
        return loads(encoded)
    else:
        return loads(f"[{encoded}]")[0]


class ParseError(BaseException):
    """Called when there is an error during parsing."""

    pass


class PersistentCustomID:
    """
    The Persistence custom_id parser.

    Used both internally and externally to parse custom_ids. Make sure to convert to a string when used in a component or modal like so: `str(your_custom_id)`.

    Attributes:
        cipher (Cipher): The cipher to use.
        tag (str): The tag to identify the component or modal.
        package (dict, list, str, int, float): The package of the component or modal.
    """

    def __init__(
        self,
        cipher: Union["Persistence", Client, Cipher],
        tag: str,
        package: Union[dict, list, str, int, float],
    ):
        """
        The constructor for the Persistence custom_id parser.

        Parameters:
            cipher (Cipher): The cipher to use.
            tag (str): The tag to identify the component or modal.
            package (dict, list, str, int, float): The package of the component or modal.
        """
        if isinstance(cipher, Client):
            self.cipher = cipher.persistence._cipher
        elif isinstance(cipher, Cipher):
            self.cipher = cipher
        else:
            try:
                self.cipher = cipher._cipher
            except AttributeError as e:
                raise ParseError("Invalid cipher provided.") from e
        self.tag = tag
        self.package = package
        if len(self.encrypt()) > 100:
            raise ParseError("Encoded custom_id is too long.")

    def encrypt(self):
        """
        Encrypts the package.

        Returns:
            str: The encrypted package.
        """
        sep = len(self.tag)
        raw = self.cipher.encrypt(self.tag+pack(self.package))
        return f"p0~{raw[:sep]}~{raw[sep:]}"

    def __str__(self):
        """Returns the encrypted custom_id as a string."""
        return self.encrypt()

    @classmethod
    def from_discord(cls, cipher: Union["Persistence", Client, Cipher], custom_id: str):
        """
        The method used internally to parse custom_ids from Discord.

        Parameters:
            cipher (Cipher): The cipher to use.
            custom_id (str): The custom_id to parse.
        """
        # if its a bot
        if hasattr(cipher, "persistence"):
            cipher = cipher.persistence._cipher
        # if its a cipher object
        elif isinstance(cipher, Cipher):
            cipher = cipher
        # if its a persistence object
        else:
            cipher = cipher._cipher
        
        _version, _tag, _payload = custom_id.split("~")

        sep = len(_tag)
        raw = cipher.decrypt(_tag+_payload, _version)
        tag = raw[:sep]
        payload = raw[sep:]
        
        
        package = unpack(payload)

        return cls(cipher, tag, package)



"""This file houses the customized Cipher class using the ff3 library."""

import logging
from random import choice, randint
from ff3 import FF3Cipher

HEX_ALPHABET = "0123456789ABCDEF"
def generate_key():
    """Generate a random key for persistence."""
    return "".join(choice(HEX_ALPHABET) for _ in range(32))

class Cipher:
    """A special Cipher object for encrypting and decrypting packages and tags."""

    def __init__(self, key=None):
        alphabet = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^*()-_=+[]{}|;':\",./<>? "
        if not key:
            logging.warning("It is highly recommended to provide a key using the cipher_key kwarg in bot.load. If you do not, a random key will be generated. Check the docs for more information.")
            key = generate_key()
            logging.info(f"Your autogenerated key is: {key}")

        tweak = "CBD09280979564" # Bad practice; should be private and random. Unfortunatly could not
        # find a way to do this. Must mention in docs to never put anything private inside of custom_ids.
        self._encrypter = FF3Cipher.withCustomAlphabet(key, tweak, alphabet)

    def encrypt(self, plain_text: str):
        """
        Encrypts a string.

        Parameters:
            plain_text (str): The string to encrypt.

        Returns:
            str: The encrypted string.
        """

        # if len(plain_text) <= 24:
        #     return self._encrypter.encrypt(plain_text)
        # chunks = [str(plain_text[i : i + 28]) for i in range(0, len(plain_text), 28)]
        # encrypted_chunks = [self._encrypter.encrypt(chunk) for chunk in chunks]

        chunk_map = list(range(0, len(plain_text), 24))
        if len(plain_text) % 24 in range(1, 4):
            chunk_map.pop()
        chunks = []
        for i in chunk_map:
            if (i/24)+1 == len(chunk_map):
                chunks.append(self._encrypter.encrypt(plain_text[i:]))
                return "".join(chunks)
            chunks.append(self._encrypter.encrypt(plain_text[i:i+24]))

    def decrypt(self, encrypted_text, version: str):
        """
        Decrypts a string.

        Parameters:
            encrypted_text (str): The string to decrypt.

        Returns:
            str: The decrypted string.
        """
        if version == "p":
            if len(encrypted_text) <= 28:
                return self._encrypter.decrypt(encrypted_text)
            chunks = [str(encrypted_text[i : i + 28]) for i in range(0, len(encrypted_text), 28)]
            decrypted_chunks = [self._encrypter.decrypt(chunk) for chunk in chunks]
            return "".join(decrypted_chunks)

        elif version == "p0":
            chunk_map = list(range(0, len(encrypted_text), 24))
            if len(encrypted_text) % 24 in range(1, 4):
                chunk_map.pop()
            chunks = []
            for i in chunk_map:
                if i/24 == len(chunk_map) - 1:
                    chunks.append(self._encrypter.decrypt(encrypted_text[i:]))
                    return "".join(chunks)
                chunks.append(self._encrypter.decrypt(encrypted_text[i:i+24]))


"""The file holding the PersistenceExension class and all of its functionality."""

from inspect import getmembers, iscoroutinefunction
from interactions import Extension, Client


class PersistenceExtension(Extension):
    """The PersistenceExtension is based off of regular Extensions, but adds callbacks for persistent components and modals"""

    def __new__(cls, client: Client, *args, **kwargs):
        """
        The extended __new__ dunder method for Persistence Extensions.

        Args:
            client (Client): An `interactions.Client` instance

        Returns:
            Extension: Returns a basic `interactions.Extension`
        """

        self = super().__new__(cls, client, *args, **kwargs)

        for _, func in getmembers(
            self, predicate=iscoroutinefunction
        ):  # credits to toricane for the inspect stuff
            if hasattr(func, "__persistence_type__"):
                if func.__persistence_type__ == "component":
                    client.persistence.component(func.__persistence_tag__)(func)
                elif func.__persistence_type__ == "modal":
                    client.persistence.modal(func.__persistence_tag__, func.__persistence_use_kwargs__)(func)


def extension_persistent_component(tag: str):
    """Callback for persistent components in extensions

    Args:
        tag (str): The tag to identify your component
    """

    def inner(coro):
        coro.__persistence_type__ = "component"
        coro.__persistence_tag__ = tag
        return coro

    return inner


def extension_persistent_modal(tag: str, use_kwargs: bool = False):
    """Callback for persistent modals in extensions

    Args:
        tag (str): The tag to identify your modal
        use_kwargs (bool): Whether to return key word arguments mapped with the custom_ids of the individual text inputs. Not recommended.
                (defaults to False)
    """

    def inner(coro):
        coro.__persistence_type__ = "modal"
        coro.__persistence_tag__ = tag
        coro.__persistence_use_kwargs__ = use_kwargs
        return coro

    return inner
